/******************************************************************************
CIF_Listener_Nexus.C

Description:

  Implementation of the Nexus Listener class.  This class is used by the
  unicast protocol to establish the server side of the connection.

CVS Information:

  $Source$
  $Date$
  $Revision$
  $State$
  $Author$
******************************************************************************/

/******************************************************************************
			     Include header files
******************************************************************************/
#include "CIF_Listener_Nexus.H"


/******************************************************************************
		       Define CIF_Listener_Nexus methods
******************************************************************************/

/*
 * CIF_Listener_Nexus::CIF_Listener_Nexus()
 */
CIF_Listener_Nexus::CIF_Listener_Nexus()
{
    nexus_mutex_init(&mutex, (nexus_mutexattr_t *) GLOBUS_NULL);

    url = CIF_NULL;
    object_state = OBJECT_STATE_INACTIVE;

    CIF::status = CIF::STATUS_SUCCESS;
}
/* CIF_Listener_Nexus::CIF_Listener_Nexus() */


CIF_Bool
CIF_Listener_Nexus::open(
    ARG_IN unsigned short		port)
{
    Nexus_Monitor		monitor(&mutex);

    /*
     * Verify that this is an inactive but valid listener object
     */
    CIF::Status			status;

    status = verify_Object_State(OBJECT_STATE_INACTIVE);
    if (status != CIF::STATUS_SUCCESS)
    {
	CIF::status = status;
	return false;
    }
    
    /*
     * Create an endpoint which will be used to receive messages from the
     * process attempting to connect to this listener
     */
    if (endpoint.activate() == false)
    {
	return false;
    }
    
    /*
     * Activate listener
     */
    if (listener.activate(endpoint, port) == false)
    {
	status = CIF::status;
	endpoint.deactivate();
	CIF::status = status;
	
	return false;
    }

    object_state = OBJECT_STATE_ACTIVE;

    return true;
}
/* CIF_Listener_Nexus::open() */


/*
 * CIF_Listener_Nexus::get_Connection()
 */
CIF_Connection *
CIF_Listener_Nexus::get_Connection(
    ARG_IN int				timeout)
{
    /*
     * Verify that this is an active listener object
     */
    {
	Nexus_Monitor			monitor(&mutex);
	CIF::Status			status;

	status = verify_Object_State(OBJECT_STATE_ACTIVE);
	if (status != CIF::STATUS_SUCCESS)
	{
	    CIF::status = status;
	    return CIF_NULL;
	}
    }

    /*
     * Allocate a new connection object and attempt to establish a connection
     */
    CIF_Connection_Nexus *		conn;

    conn = allocate_Connection();
    if (CIF::status != CIF::STATUS_SUCCESS)
    {
	return CIF_NULL;
    }

    if (conn->open(endpoint, timeout))
    {
	return (CIF_Connection *) conn;
    }
    else
    {
	delete conn;

	CIF::status = CIF::STATUS_FAILURE;
	return CIF_NULL;
    }
}
/* CIF_Listener_Nexus::get_Connection() */


/*
 * CIF_Listener_Nexus::get_URL()
 */
char *
CIF_Listener_Nexus::get_URL()
{
    Nexus_Monitor		monitor(&mutex);

    /*
     * Verify that this is a valid listener object
     */
    CIF::Status			status;

    status = verify_Object_State(OBJECT_STATE_ACTIVE);
    if (status != CIF::STATUS_SUCCESS)
    {
	CIF::status = status;
	return CIF_NULL;
    }

    if (url == CIF_NULL)
    {
	int			len;

	len = strlen(get_URL_Scheme()) + strlen(listener.get_Hostname()) + 14;
	url = new char [len];

	sprintf(url,
		"%s://%s:%d",
		get_URL_Scheme(),
		listener.get_Hostname(),
		listener.get_Port());
    }

    return url;
}


/*
 * CIF_Listener_Nexus::l_close()
 */
CIF_Bool
CIF_Listener_Nexus::l_close()
{
    CIF::Status			status;
    CIF_Bool			rc;

    /*
     * If the listener is current active, then deactivate it; otherwise return
     * an error
     */
    status = verify_Object_State(OBJECT_STATE_ACTIVE);
    if (status == CIF::STATUS_SUCCESS)
    {
        listener.deactivate();
	endpoint.deactivate();

	object_state = OBJECT_STATE_INACTIVE;
	rc = true;
    }
    else

    {
	rc = false;
    }

    CIF::status = status;
    return rc;
}
/* CIF_Listener_Nexus::l_close() */


/*
 * CIF_Listener_Nexus::close()
 */
CIF_Bool
CIF_Listener_Nexus::close()
{
    Nexus_Monitor		monitor(&mutex);

    return l_close();
}

/*
 * CIF_Listener_Nexus::~CIF_Listener_Nexus()
 */
CIF_Listener_Nexus::~CIF_Listener_Nexus()
{
    {
	Nexus_Monitor		monitor(&mutex);

	l_close();
	object_state = OBJECT_STATE_DELETED;
    }
    
    nexus_mutex_destroy(&mutex);
    if (url != CIF_NULL)
    {
	delete [] url;
    }
}
/* CIF_Listener_Nexus::~CIF_Listener_Nexus() */


/*
 * CIF_Listener::verify_Object_State()
 */
CIF::Status
CIF_Listener_Nexus::verify_Object_State(
    Object_State			desired_state)
{
    if (object_state == desired_state)
    {
	return CIF::STATUS_SUCCESS;
    }
    
    /*
     * We aren't in the state we explected to be, so return an appropriate
     * error to the user
     */
    switch (object_state)
    {
      case OBJECT_STATE_DELETED:
	NexusAssert(object_state != OBJECT_STATE_DELETED);
	return CIF::STATUS_OBJECT_DELETED;
	    
      case OBJECT_STATE_ACTIVE:
	return CIF::STATUS_OBJECT_ACTIVE;
	
      case OBJECT_STATE_INACTIVE:
	return CIF::STATUS_OBJECT_INACTIVE;

      default:
	return CIF::STATUS_INTERNAL_ERROR;
    }
}
/* CIF_Endpoint::verify_Object_State() */

